﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace XyDraw_NS
{
    /// <summary>
    /// UI Component to draw vertical or horizontal axis, with tic marks and labels.
    /// Settings can be altered using a GridProperty.
    /// 
    /// Author: Dennis Lang 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// </summary>
    public partial class GraphAxis : UserControl
    {
        public GraphAxis()
        {
            InitializeComponent();
            DefaultSettings();
        }

        public enum Direction { Vertical, Horizontal };
        public enum Layout { None, LeftBottom, Center, Zero };

        [Serializable]
        public class GraphAxisSettings
        {
            public GraphAxisSettings()
            {
                direction = Direction.Vertical;
                layout = Layout.LeftBottom;
                lineColor = Color.Black;
                lineWidth = 3;
                ticColor = Color.Black;
                ticWidth = 3;
                textColor = Color.Black;
                textFont = new Font("Calibri", 20f);
                origin = 0;
                drawTic = true;
                ticLen = 10;
            }

            public void Build()
            {
            }

            Direction direction;
            Layout layout;

            string label;
            Font textFont;
            Color textColor;

            float origin;
            bool drawTic;
            float ticLen;

            Color lineColor;
            float lineWidth;
            Color ticColor;
            float ticWidth;

            [CategoryAttribute("General"),
            DescriptionAttribute("Direction")]
            public Direction AxisDirection
            {
                get { return direction; }
                set { direction = value; }
            }

            [CategoryAttribute("General"),
            DescriptionAttribute("Layout")]
            public Layout AxisLayout
            {
                get { return layout; }
                set { layout = value; }
            }

            [CategoryAttribute("General"),
            DescriptionAttribute("Draw Tic")]
            public bool DrawTic
            {
                get { return drawTic; }
                set { drawTic = value; }
            }

            [CategoryAttribute("General"),
            DescriptionAttribute("Line Width")]
            public float LineWidth
            {
                get { return lineWidth; }
                set { lineWidth = value; }
            }

            [CategoryAttribute("General"),
            DescriptionAttribute("Tic Width")]
            public float TicWidth
            {
                get { return ticWidth; }
                set { ticWidth = value; }
            }

            [CategoryAttribute("General"),
            DescriptionAttribute("Tic Length")]
            public float TicLength
            {
                get { return ticLen; }
                set { ticLen = value; }
            }

            [CategoryAttribute("General"),
            DescriptionAttribute("Origin")]
            public float Origin
            {
                get { return origin; }
                set { origin = value; }
            }

            [CategoryAttribute("Colors"),
            Browsable(false),
            DescriptionAttribute("Line Pen")]
            public Pen LinePen
            {
                get { return new Pen(lineColor, lineWidth); }
            }

            [CategoryAttribute("Colors"),
            DescriptionAttribute("Line Color")]
            public Color LineColor
            {
                get { return lineColor; }
                set { lineColor = value; }
            }

            [CategoryAttribute("Colors"),
            Browsable(false),
            DescriptionAttribute("Tic Pen")]
            public Pen TicPen
            {
                get { return new Pen(ticColor, ticWidth); }
            }

            [CategoryAttribute("Colors"),
            DescriptionAttribute("Tic Color")]
            public Color TicColor
            {
                get { return ticColor; }
                set { ticColor = value; }
            }

            [CategoryAttribute("Colors"),
            DescriptionAttribute("Text Color")]
            public Color TextColor
            {
                get { return textColor; }
                set { textColor = value; }
            }

            [CategoryAttribute("Text"),
            DescriptionAttribute("Label")]
            public string Label
            {
                get { return label; }
                set { label = value; }
            }

            [CategoryAttribute("Text"),
            DescriptionAttribute("Text Font")]
            public Font TextFont
            {
                get { return textFont; }
                set { textFont = value; }
            }

        };

        GraphAxisSettings s;

        public object Settings
        {
            get { return s; }
            set { s = value as GraphAxisSettings; }
        }

        public GraphAxisSettings AxisSettings
        {
            get { return s; }
            set { s = value; Refresh(); }
        }

        public void DefaultSettings()
        {
            s = new GraphAxisSettings();
            s.TextFont = new Font("Calibri", 20f);
        }

        private void Refresh()
        {
        }

        const int gap = 2;
        public void DrawOn(Image image, RectangleF bounds)
        {
            switch (s.AxisLayout)
            {
                case Layout.None:
                    break;
                case Layout.LeftBottom:
                    Graphics g = Graphics.FromImage(image);
                    SizeF txtSiz = g.MeasureString("-222.22", s.TextFont);
                    s.Origin = Math.Max(txtSiz.Width, txtSiz.Height) + s.LineWidth + gap;
                    DrawAxisOn(image, bounds);
                    break;
                case Layout.Center:
                    s.Origin = (s.AxisDirection == Direction.Vertical) ? image.Width / 2 : image.Height / 2;
                    DrawAxisOn(image, bounds);
                    break;
                case Layout.Zero:
                    break;
            }
        }

#if false
        PointF Proj(PointF pt)
        {
            return new PointF((pt.X - m_bounds.X) * m_scale.X,
                pictureBox.BackgroundImage.Height - (pt.Y - m_bounds.Y) * m_scale.Y);
        }

        PointF RevProj(PointF screen)
        {
            return new PointF(screen.X / m_screenScale.X + m_bounds.X,
                (pictureBox.BackgroundImage.Height - screen.Y) / m_screenScale.Y + m_bounds.Y);
        }
#endif

        public void DrawAxisOn(Image image, RectangleF bounds)
        {
            Graphics g = Graphics.FromImage(image);
            float off = 10;
            float org = s.Origin;
            float ticLen = s.TicLength;
            float rate = 0.10f;  // 10 percent
            SolidBrush textBrush = new SolidBrush(s.TextColor);

            float scaleX = bounds.Width / image.Width;
            float scaleY = bounds.Height / image.Height;

            if (s.AxisDirection == Direction.Vertical)
            {
                // Vertical Axis
                PointF pt1 = new PointF(org, off);
                PointF pt2 = new PointF(pt1.X, image.Height - pt1.Y);
                g.DrawLine(s.LinePen, pt1, pt2);

                for (float f = rate; f < 1.0; f += rate)
                {
                    float ticP = pt1.Y + (pt2.Y - pt1.Y) * f;
                    g.DrawLine(s.TicPen, pt1.X, ticP, pt1.X + ticLen, ticP);

                    string lbl = (bounds.Y + bounds.Height - ticP * scaleY).ToString("F2");
                    SizeF txtSiz = g.MeasureString(lbl, s.TextFont);
                    g.DrawString(lbl, s.TextFont, textBrush, new PointF(pt1.X - txtSiz.Width - gap, ticP - txtSiz.Height / 2));
                }
            }
            else
            {
                // Horizontal Axis
                PointF pt1 = new PointF(off, image.Height - org);
                PointF pt2 = new PointF(image.Width - pt1.X, pt1.Y);
                g.DrawLine(s.LinePen, pt1, pt2);

                for (float f = rate; f < 1; f += rate)
                {
                    float ticP = pt1.X + (pt2.X - pt1.X) * f;
                    g.DrawLine(s.TicPen, ticP, pt1.Y, ticP, pt1.Y - ticLen);

                    string lbl = (bounds.X + ticP * scaleX).ToString("F2");
                    SizeF txtSiz = g.MeasureString(lbl, s.TextFont);
                    g.DrawString(lbl, s.TextFont, textBrush, new PointF(ticP - txtSiz.Width / 2, pt1.Y + gap));

                }
            }
        }
    }
}
