﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;

using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace XyDraw_NS
{
    /// <summary>
    /// Simple program to draw XY data series
    /// 
    /// Author:  DLang 2010 
    /// http://home.comcast.net/~lang.dennis/index.html
    /// </summary>
    public partial class XyDraw : Form
    {
        public XyDraw()
        {
            InitializeComponent();

            this.Text = this.titleBtn.Text = ProductNameAndVersion();
            
            vAxis.DefaultSettings();
            hAxis.DefaultSettings();

            pictureBox.BackgroundImageLayout = ImageLayout.Zoom;
            pictureBox.SizeMode = PictureBoxSizeMode.Zoom;

            pictureBox.Image = new Bitmap(800, 800);
            Graphics g = Graphics.FromImage(pictureBox.Image);
            g.Clear(Color.Transparent);

            settings.item.bgColor = colorBtn.Color = Color.FromArgb(255, 200, 200, 210);
            colorBtn.changed += new EventHandler(ColorChanged);

            RestoreSettings();

            fileView.ListViewItemSorter = new ListViewColumnSorter(ListViewColumnSorter.SortDataType.eAuto);
        }   

        #region ==== Data Members

        internal class Points
        {
            public Pen pen;
            public RectangleF bounds;
            public string filename;
            public List<PointF> points = new List<PointF>();
        };

        List<Points> filePoints = new List<Points>();
        List<int>    selectedFiles = new List<int>();

        public GraphAxis vAxis = new GraphAxis();
        public GraphAxis hAxis = new GraphAxis();

        RectangleF  m_bounds;
        PointF      m_scale;
        PointF      m_screenScale;
        PrintIt printIt = new PrintIt();

    
        [Serializable]
        public class SettingItem : ICloneable
        {
            public SettingItem()
            {
                vAxisSettings = new GraphAxis.GraphAxisSettings();
                hAxisSettings = new GraphAxis.GraphAxisSettings();

                colors = new Color[] { Color.Red, Color.Blue, Color.Green, Color.Yellow, Color.Orange, 
                    Color.Olive, Color.Brown, Color.Salmon};

                size = new Size(400, 400);

                bgColor = Color.DarkBlue;
            }

            public string name;
            public string desc;
            public Size size;
            public Color bgColor;
            public Color[] colors;

            public GraphAxis.GraphAxisSettings vAxisSettings;
            public GraphAxis.GraphAxisSettings hAxisSettings;

            public PointF       scale;
            public PointF       screenScale;
            public RectangleF   bounds;

            public List<string> selectedFiles;
            public List<Color>  selectedColors;

            // deep copy in separeate memory space
            public object Clone()
            {
                MemoryStream ms = new MemoryStream();
                BinaryFormatter bf = new BinaryFormatter();
                bf.Serialize(ms, this);
                ms.Position = 0;
                object obj = bf.Deserialize(ms);
                ms.Close();
                return obj;
            }

        };

        [Serializable]
        public class Settings
        {
            public Settings()
            {
                item = new SettingItem();
            }

            public SettingItem item;
            public List<SettingItem> bookmarks;
        }

        public Settings settings = new Settings();
        const string SettingsFilename = @"\XyDraw.dat";

        #endregion

        #region ==== Basic UI controls 

        public static string ProductNameAndVersion()
        {
            string appName = Application.ProductName;
            string appVern = Application.ProductVersion;
            return appName + " v" + appVern.Substring(0, 3); //  Get part of versoin string "n.n"
        }

        private void closeBtn_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            SaveSettings();
            base.OnClosing(e);
        }

        Image bgImage = null;
        protected override void OnResizeBegin(EventArgs e)
        {
            bgImage = this.BackgroundImage;
            this.BackgroundImage = null;
            base.OnResizeBegin(e);
        }

        protected override void OnResizeEnd(EventArgs e)
        {
            base.OnResizeEnd(e);
            this.BackgroundImage = bgImage;
        }

        private void about_Click(object sender, EventArgs e)
        {
            About about = new About();
            about.Show();
        }

        #endregion

        #region ==== Settings

        private string SettingsFile()
        {
            string appPath = System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
            string ourFile = appPath + SettingsFilename;
            return ourFile;
        }

        private void SaveSettings()
        {
            string ourFile = SettingsFile();

            settings.item.size = this.Size;
            settings.item.bgColor = colorBtn.Color;

            settings.item.vAxisSettings = vAxis.AxisSettings;
            settings.item.hAxisSettings = hAxis.AxisSettings;

            settings.item.bounds = m_bounds;
            settings.item.scale = m_scale;
            settings.item.screenScale = m_screenScale;
            settings.item.selectedColors = new List<Color>();
            settings.item.selectedFiles = new List<string>();

            foreach (int fileIdx in this.selectedFiles)
            {
                settings.item.selectedFiles.Add(filePoints[fileIdx].filename);
                settings.item.selectedColors.Add(filePoints[fileIdx].pen.Color);
            }

            try
            {
                Stream stream = File.OpenWrite(ourFile);
                BinaryFormatter bFormatter = new BinaryFormatter();
                bFormatter.Serialize(stream, settings);
                stream.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        public void RestoreSettings()
        {
            string ourFile = SettingsFile();
            settings.item.size = this.Size;

            try
            {
                Stream stream = File.OpenRead(ourFile);
                try
                {
                    BinaryFormatter bFormatter = new BinaryFormatter();
                    Settings newSettings = (Settings)bFormatter.Deserialize(stream);
                    settings = newSettings;
                }
                catch (Exception ex)
                {
                    MessageBox.Show("Sorry saved settings reset");
                }
                stream.Close();
            }
            catch (Exception ex)
            {
                // MessageBox.Show(ex.Message);
                return;
            }

            // Load active setting item.
            LoadSettingItem();

            // Load Bookmark View
            this.bookmarkView.Items.Clear();
            if (this.settings.bookmarks != null)
            {
                foreach (SettingItem sItem in this.settings.bookmarks)
                {
                    ListViewItem bmItem = bookmarkView.Items.Add(sItem.name);
                    bmItem.SubItems.Add(sItem.desc);
                }
            }
        }

        private void LoadSettingItem()
        {
            doDraw = false;

            if (settings.item == null)
            {
                settings.item = new SettingItem();
                settings.item.vAxisSettings = vAxis.AxisSettings;
                settings.item.hAxisSettings = hAxis.AxisSettings;
            }

            this.Size = settings.item.size;
            this.colorBtn.Color = settings.item.bgColor;

            this.vAxis.AxisSettings = settings.item.vAxisSettings;
            this.hAxis.AxisSettings = settings.item.hAxisSettings;


            // Populate Color View
            colorList.Items.Clear();
            foreach (Color color in settings.item.colors)
            {
                ListViewItem clrItem =
                    colorList.Items.Add((colorList.Items.Count + 1).ToString());
                clrItem.UseItemStyleForSubItems = false;
                clrItem.SubItems.Add("").BackColor = color;
            }

            // Update axis settings. 
            vAxis.AxisSettings.AxisDirection = GraphAxis.Direction.Vertical;
            vAxisPropGrid.SelectedObject = vAxis.Settings;

            hAxis.AxisSettings.AxisDirection = GraphAxis.Direction.Horizontal;
            hAxisPropGrid.SelectedObject = hAxis.Settings;

            SetAxisCheck(vAxis.AxisSettings.AxisLayout);

            fileView.Items.Clear();
            filePoints.Clear();

            if (settings.item.selectedFiles != null)
            {
                int cnt = settings.item.selectedFiles.Count;
                int fileIdx = 0; 
                while (fileIdx < cnt)
                {
                    if (LoadFile(settings.item.selectedFiles[fileIdx], fileIdx))
                    {
                        if (settings.item.selectedColors != null && fileIdx < settings.item.selectedColors.Count)
                        {
                            fileView.Items[fileIdx].UseItemStyleForSubItems = false;
                            fileView.Items[fileIdx].SubItems[2].Font = new Font("Marlett", this.Font.Size);
                            fileView.Items[fileIdx].SubItems[2].Text = "gggg";
                            fileView.Items[fileIdx].SubItems[2].BackColor =
                                fileView.Items[fileIdx].SubItems[2].ForeColor = 
                                filePoints[fileIdx].pen.Color =
                                settings.item.selectedColors[fileIdx];
                        }
                        fileView.Items[fileIdx].Selected = true;
                            fileIdx++;
                    }
                    else
                    {
                        settings.item.selectedFiles.RemoveAt(fileIdx);
                        cnt--;
                    }
                }

                m_bounds = settings.item.bounds;
                m_screenScale = settings.item.screenScale;
                m_scale = settings.item.scale;
            }

            doDraw = true;
            if (fileView.SelectedItems.Count > 0)
                DrawPoints();
        }

        #endregion

        #region ==== File View and File Load
        private void loadBtn_Click(object sender, EventArgs e)
        {
            pictureBox.BackgroundImage = new Bitmap(800, 800);
            pictureBox.BackColor = this.colorBtn.Color;

            openFileDialog.FileName = fileBox.Text;
            if (openFileDialog.ShowDialog() == DialogResult.OK)
            {
                doDraw = false;
                LoadFiles(openFileDialog.FileNames);
                doDraw = true;
            }

            if (fileView.CheckedItems.Count == 0 && fileView.Items.Count > 0)
                fileView.Items[0].Checked = true;
        }

        private void fileBox_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (e.KeyCode == Keys.Tab)
                e.IsInputKey = true;
        }

        private void fileBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                pictureBox.BackgroundImage = new Bitmap(800, 800);
                pictureBox.BackColor = this.colorBtn.Color;

                // TODO - check for duplicates
                int fileIdx = filePoints.Count;
                LoadFile(fileBox.Text, fileIdx);

                if (fileView.CheckedItems.Count == 0 && fileView.Items.Count > 0)
                    fileView.Items[0].Checked = true;
            }
        }

        private void AddToFileView(int fileIdx)
        {
            Points filePts = filePoints[fileIdx];

            ListViewItem fileItem = fileView.Items.Add(filePts.filename);
            fileItem.Tag = fileIdx;
            fileItem.UseItemStyleForSubItems = false;
            fileItem.SubItems.Add(Path.GetFileName(filePts.filename));
            fileItem.SubItems.Add("").BackColor = filePts.pen.Color;
            fileItem.SubItems.Add(filePts.points.Count.ToString("#,##0"));
            fileItem.SubItems.Add(filePts.bounds.X.ToString("F3"));
            fileItem.SubItems.Add(filePts.bounds.Y.ToString("F3"));
            double area = filePts.bounds.Size.Width * filePts.bounds.Size.Height;
            fileItem.SubItems.Add(area.ToString("F3"));

            fileItem.SubItems[2].ForeColor = filePts.pen.Color;
            fileItem.SubItems[2].Font = new Font("Marlett", this.Font.Size);
            fileItem.SubItems[2].Text = "gggg";
        }

        private void progressBar_Click(object sender, EventArgs e)
        {
            progressBar.Visible = false;
        }

        private void LoadFiles(string[] filepaths)
        {
            // fileView.Items.Clear();
            if (filePoints == null)
                filePoints = new List<Points>();
            progressBar.Value = 0;
            progressBar.Visible = true;

            int fileCnt = filepaths.Length;
            progressBar.Maximum = fileCnt;

            for (int fileIdx = 0; fileIdx < fileCnt && progressBar.Visible; fileIdx++)
            {
                string filepath = filepaths[fileIdx];
                LoadFile(filepath, filePoints.Count);
                progressBar.Increment(1);
            }

            progressBar.Visible = false;
        }

        private bool LoadFile(string filepath, int fileIdx)
        {
            fileBox.Text = Path.GetFileName(filepath);

            while (filePoints.Count <= fileIdx)
                filePoints.Add(new Points());

            filePoints[fileIdx].points.Clear();
            filePoints[fileIdx].filename = filepath;
            filePoints[fileIdx].pen = new Pen(settings.item.colors[fileIdx % settings.item.colors.Length]);

            try
            {
                TextReader txtReader = new StreamReader(filepath);

                string txtLine;
                if (yxImportCk.Checked)
                {
                    while ((txtLine = txtReader.ReadLine()) != null)
                    {
                        string[] fields = txtLine.Trim().Split(',');
                        if (fields.Length > 1)
                        {
                            try
                            {
                                PointF pt = new PointF(0, 0);
                                pt.Y = float.Parse(fields[0]);
                                pt.X = float.Parse(fields[1]);
                                filePoints[fileIdx].points.Add(pt);
                            }
                            catch { }
                        }
                    }
                }
                else
                {
                    while ((txtLine = txtReader.ReadLine()) != null)
                    {
                        string[] fields = txtLine.Trim().Split(',');
                        if (fields.Length > 1)
                        {
                            try
                            {
                                PointF pt = new PointF(0, 0);
                                pt.X = float.Parse(fields[0]);
                                pt.Y = float.Parse(fields[1]);
                                filePoints[fileIdx].points.Add(pt);
                            }
                            catch { }
                        }
                    }
                }

                filePoints[fileIdx].bounds = ComputeBounds(fileIdx);
                status.Text = string.Format("Loaded {0} points, {1:F2},{2:F2} {3:F2},{4:F2}",
                    filePoints[fileIdx].points.Count,
                    filePoints[fileIdx].bounds.X, filePoints[fileIdx].bounds.Y,
                    filePoints[fileIdx].bounds.Right, m_bounds.Bottom);

                AddToFileView(fileIdx);
                return true;
            }
            catch 
            {
                return false;
            }
        }

        bool doFileViewChecked = true;
        private void fileView_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            if (doFileViewChecked)
            {
                UpdatePointView();
            }
        }

        private void UpdatePointView()
        {
            listView.Items.Clear();
            selectedFiles.Clear();

            foreach (ListViewItem fileItem in fileView.CheckedItems)
            {
                int fileIdx = (int)fileItem.Tag;
                selectedFiles.Add(fileIdx);
                ViewPoints(fileIdx);
            }

            ComputeSelectedBounds();
            DrawPoints();
        }

        private void fileView_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.Control)
            {
                switch (e.KeyCode)
                {
                    case Keys.A:
                        foreach (ListViewItem fileItem in fileView.Items)
                        {
                            fileItem.Selected = true;
                        }
                        break;
                }

                return;
            }


            switch (e.KeyCode)
            {
                // Toggle selected
                case Keys.Space:
                    if (fileView.SelectedItems.Count > 0)
                    {
                        bool isChecked = fileView.SelectedItems[0].Checked;
                        foreach (ListViewItem fileItem in fileView.SelectedItems)
                        {
                            fileItem.Checked = isChecked;
                        }
                    }
                    break;
                // Clear all
                case Keys.OemMinus:
                case Keys.D0:
                case Keys.C:
                    doFileViewChecked = false;
                    foreach (ListViewItem fileItem in fileView.Items)
                    {
                        fileItem.Checked = false;
                    }
                    doFileViewChecked = true;
                    UpdatePointView();
                    break;
                // Set all
                case Keys.Oemplus:
                case Keys.D1:
                case Keys.S:
                    doFileViewChecked = false;
                    progressBar.Value = 0;
                    progressBar.Visible = true;
                    progressBar.Maximum = fileView.Items.Count;

                    foreach (ListViewItem fileItem in fileView.Items)
                    {
                        if (progressBar.Visible)
                            fileItem.Checked = true;

                        progressBar.Increment(1);
                    }
                    progressBar.Visible = false;
                    doFileViewChecked = true;
                    UpdatePointView();
                    break;
            }
        }

        private void fileView_MouseUp(object sender, MouseEventArgs e)
        {
            ListViewHitTestInfo hitInfo = fileView.HitTest(e.Location);
            if (hitInfo != null && hitInfo.Item != null)
            {
                if (e.Button == MouseButtons.Left)
                {
                    hitInfo.Item.Checked = !hitInfo.Item.Checked;
                }
                else
                {
                    hitInfo.Item.Selected = true;
                }
            }

        }

        #region ==== File Context Menu
        private void removeFileMenu_Click(object sender, EventArgs e)
        {
            doDraw = false;
            int cnt = fileView.Items.Count;
            for (int idx = cnt - 1; idx >= 0; idx--)
            {
                ListViewItem fileItem = fileView.Items[idx];
                if (fileItem.Checked == false)
                {
                    this.filePoints[idx] = new Points();  // clear position, don't remove it.
                    fileView.Items.RemoveAt(idx);
                }
            }
            doDraw = true;
        }
        private void clearFileMenu_Click(object sender, EventArgs e)
        {
            fileView.Items.Clear();
            selectedFiles.Clear();
            filePoints.Clear();
            DrawPoints();
        }

        private void openFolderMenu_Click(object sender, EventArgs e)
        {
            if (fileView.SelectedItems.Count > 0)
            {
                LaunchExplorer(fileView.SelectedItems[0].Text);
            }
        }

        private void Launch(string filepath)
        {
            System.Environment.CurrentDirectory = Path.GetDirectoryName(filepath);
            System.Diagnostics.Process.Start(filepath);
        }

        private void LaunchExplorer(string filepath)
        {
            string dir = Path.GetDirectoryName(filepath);
            System.Diagnostics.Process.Start("explorer.exe", dir);
        }

        private void LaunchCmd(string filepath)
        {
            string dir = Path.GetDirectoryName(filepath);
            System.Diagnostics.Process.Start("cmd.exe", "/k cd \"" + dir + "\"");
        }

        #endregion

        #endregion

        #region ==== Bounding box and Point Projection

        private void resetBtn_Click(object sender, EventArgs e)
        {
            ComputeSelectedBounds();
            DrawPoints();
        }

        private void ComputeSelectedBounds()
        {
            m_bounds = RectangleF.Empty;
            if (selectedFiles.Count > 0)
                m_bounds = filePoints[selectedFiles[0]].bounds;

            foreach (int fileIdx in selectedFiles)
            {
                m_bounds = RectangleF.Union(m_bounds, filePoints[fileIdx].bounds);
            }

            float inflateRate = 10.0f;
            m_bounds.Inflate(new SizeF(m_bounds.Size.Width / inflateRate, 
                    m_bounds.Size.Height / inflateRate));
            if (aspectCk.Checked)
            {
                float maxDim = Math.Max(m_bounds.Width, m_bounds.Height);
                m_bounds.Size = new SizeF(maxDim, maxDim);
            }
        }

        private RectangleF ComputeBounds(int fileIdx)
        {
            float minX = float.MaxValue;
            float minY = float.MaxValue;
            float maxX = float.MinValue;
            float maxY = float.MinValue;

            foreach (PointF pt in filePoints[fileIdx].points)
            {
                minX = Math.Min(minX, pt.X);
                maxX = Math.Max(maxX, pt.X);
                minY = Math.Min(minY, pt.Y);
                maxY = Math.Max(maxY, pt.Y);
            }

            return new RectangleF(minX, minY, maxX - minX, maxY - minY);
        }

        PointF Proj(PointF pt)
        {
            return new PointF((pt.X - m_bounds.X) * m_scale.X,
                pictureBox.BackgroundImage.Height - (pt.Y - m_bounds.Y) * m_scale.Y);
        }

        PointF RevProj(PointF screen)
        {
            // Bounds upper left corner will shift if image is not full screen.
            if (pictureBox.Width > pictureBox.Height)
                screen.X -= (pictureBox.Width - pictureBox.Height)/2f;
            else
                screen.Y -= (pictureBox.Height - pictureBox.Width) / 2f;

            return new PointF(screen.X * m_screenScale.X + m_bounds.X,
                    m_bounds.Y + m_bounds.Height - screen.Y * m_screenScale.Y);
        }

        #endregion

        #region ==== View Points
        private void ViewPoints(int fileIdx)
        {
            if (pointViewCk.Checked)
            {
                string filename = fileView.Items[fileIdx].Text;
                Color color = fileView.Items[fileIdx].SubItems[2].BackColor;

                // listView.Items.Clear();
                listView.BeginUpdate();
                foreach (PointF pt in filePoints[fileIdx].points)
                {
                    ListViewItem ptItem = listView.Items.Add((listView.Items.Count + 1).ToString());
                    ptItem.UseItemStyleForSubItems = false;

                    ptItem.SubItems.Add(pt.X.ToString());
                    ptItem.SubItems.Add(pt.Y.ToString());
                    ptItem.SubItems.Add(fileIdx.ToString()).ForeColor = color;
                    ptItem.SubItems.Add(filename);
                    ptItem.Tag = pt;

                    if (listView.Items.Count > 4000)
                    {
                        listView.Items.Add("***").SubItems.Add("Too Many");
                        break;
                    }
                }

                if (listView.Items.Count > 0)
                    listView.AutoResizeColumns(ColumnHeaderAutoResizeStyle.ColumnContent);

                listView.EndUpdate();
            }
        }

        private void FindPoint(PointF findPt)
        {
#if false
            double minDst = double.MaxValue;
            int idx = -1;

            for (int ptIdx = 0; ptIdx < points.Count; ptIdx++)
            {
                PointF pt = (points[ptIdx]);
                double dx = pt.X - findPt.X;
                double dy = pt.Y - findPt.Y;
                double dst = dx * dx + dy * dy;
                if (dst < minDst)
                {
                    idx = ptIdx;
                    minDst = dst;
                }
            }

            listView.SelectedItems.Clear();
            if (idx != -1)
            {
                listView.Items[idx].Selected = true;
                listView.Items[idx].EnsureVisible();
            }
#endif
        }

        private void listView_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            DrawMarkers();

        }

        private void DrawMarkers()
        {
            Graphics g = Graphics.FromImage(pictureBox.Image);
            g.Clear(Color.Transparent);
            Pen pen = Pens.Green;
            Brush brush = Brushes.Green;

            int hsz = markerSize.Value;
            int sz = hsz * 2;

            foreach (ListViewItem ptItem in listView.SelectedItems)
            {
                PointF ptf = (PointF)ptItem.Tag;
                PointF ptImg = Proj(ptf);
                g.FillEllipse(brush, ptImg.X - hsz, ptImg.Y - hsz, sz, sz);

                if (ptItem.Index > 0 && listView.Items[ptItem.Index - 1].Selected)
                {
                    PointF ptfPrev = (PointF)listView.Items[ptItem.Index - 1].Tag;
                    PointF ptImgPrev = Proj(ptfPrev);
                    g.DrawLine(pen, ptImgPrev, ptImg);
                }
            }

            pictureBox.Refresh();
        }

        #endregion

        #region ==== Draw Points and Picture Box

        bool doDraw = true;
        private void DrawPoints()
        {
            if (doDraw == false) 
                return;

            if (selectedFiles.Count == 0)
            {
                System.ComponentModel.ComponentResourceManager resources = 
                    new System.ComponentModel.ComponentResourceManager(typeof(XyDraw));
                this.pictureBox.BackgroundImage = ((System.Drawing.Image)(resources.GetObject(
                    "pictureBox.BackgroundImage")));
                return;
            }

            float scaleX = pictureBox.BackgroundImage.Width / m_bounds.Width;
            float scaleY = pictureBox.BackgroundImage.Height / m_bounds.Height;
            m_scale = new PointF(scaleX, scaleY);

            // Source image is square, pictureBox is in zoom mode (fixed aspect), so 
            // scale must be square.
            int minDim = Math.Min(pictureBox.Width, pictureBox.Height);
            m_screenScale = new PointF(m_bounds.Width / minDim,
                    m_bounds.Height / minDim);

            Graphics g = Graphics.FromImage(pictureBox.BackgroundImage);
            // g.Clear(Color.Transparent);
            g.Clear(this.colorBtn.Color);
            g.Dispose();

            int alpha = alphaBar.Value;

            foreach (int fileIdx in selectedFiles)
            {
                Pen pen = (lineWidth.Value > 0) ? new Pen(filePoints[fileIdx].pen.Color, lineWidth.Value) : null;
                Brush brush = (alpha > 0) ? new SolidBrush(Color.FromArgb(alpha, filePoints[fileIdx].pen.Color)) : null;
                DrawPoints(filePoints[fileIdx].points, pen, brush);
            }

            g = Graphics.FromImage(pictureBox.Image);
            g.Clear(Color.Transparent);

            vAxis.DrawOn(pictureBox.BackgroundImage, m_bounds);
            hAxis.DrawOn(pictureBox.BackgroundImage, m_bounds);

            pictureBox.Refresh();
        }

        private void DrawPoints(List<PointF> points, Pen pen, Brush brush)
        {
            Graphics g = Graphics.FromImage(pictureBox.BackgroundImage);

            PointF[] polyPoints = new PointF[points.Count];
            for (int ptIdx = 0; ptIdx < points.Count; ptIdx++)
            {
                polyPoints[ptIdx] = Proj(points[ptIdx]);
            }

            if (brush != null && points.Count > 2 && polyPoints[0] == polyPoints[points.Count -1])
                g.FillPolygon(brush, polyPoints);
            if (pen != null)
                g.DrawPolygon(pen, polyPoints);
        }

        private void pictureBox_Resize(object sender, EventArgs e)
        {
            DrawPoints();
        }

        private void aspectCk_CheckedChanged(object sender, EventArgs e)
        {
            if (aspectCk.Checked)
            {
                float maxDim = Math.Max(m_bounds.Width, m_bounds.Height);
                m_bounds.Size = new SizeF(maxDim, maxDim);
            }
            else
            {
                ComputeSelectedBounds();
            }
            this.DrawPoints();
        }

        Point lastPt = Point.Empty;
        private void pictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            if (lastPt != Point.Empty)
            {
                int deltaX = lastPt.X - e.X;
                int deltaY = lastPt.Y - e.Y;
                PointF pt = RevProj(e.Location);
                status.Text = pt.ToString();

                if (e.Button == MouseButtons.Right)
                {
                    // Zoom
                    int zoomLvl = (Math.Abs(deltaX) > Math.Abs(deltaY)) ? deltaX : deltaY;
                    float zoomInc = zoomLvl / 100.0f;
                    // m_bounds.Size += new SizeF(zoomInc, zoomInc);
                    m_bounds.Inflate(zoomInc, zoomInc);
                    DrawPoints();
                }
                else if (e.Button == MouseButtons.Left)
                {
                    PointF pt1 = RevProj(lastPt);
                    PointF pt2 = pt;
                    m_bounds.X += (pt1.X - pt2.X);
                    m_bounds.Y += (pt1.Y - pt2.Y);
                    DrawPoints();
                }
                else if (System.Windows.Forms.Control.ModifierKeys == Keys.Shift)
                {
                    FindPoint(pt);
                }
            }

            lastPt = e.Location;
        }

        #endregion

        #region ==== Auxilary UI actions (print and save)
        private void printSetupMenu_Click(object sender, EventArgs e)
        {
            printIt.PageSetup();
        }

        private void printMenu_Click(object sender, EventArgs e)
        {
            printIt.Print(this, new Rectangle(this.Location, this.Size), this.Text);
        }

        private void saveImageAsMenu_Click(object sender, EventArgs e)
        {
            Bitmap image = printIt.GrabScreenImage(new Rectangle(this.Location, this.Size));

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                image.Save(saveFileDialog.FileName);
            }
        }

        private void saveGraphImageAsMenu_Click(object sender, EventArgs e)
        {
            Point corner = this.pictureBox.PointToScreen(Point.Empty);
            Bitmap image = printIt.GrabScreenImage(new Rectangle(corner, pictureBox.Size));

            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                image.Save(saveFileDialog.FileName);
            }
        }

        private void deleteSettingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string filename = SettingsFile();
            try
            {
                File.Delete(filename);
            }
            catch { }
            LaunchExplorer(filename);
        }

        private void lineWidth_Scroll(object sender, EventArgs e)
        {
            DrawPoints();
        }
        private void markerSize_Scroll(object sender, EventArgs e)
        {
            DrawMarkers();
        }

        private void ColorChanged(object sender, EventArgs e)
        {
            this.pictureBox.BackColor = colorBtn.Color;
            DrawPoints();
        }
        #endregion

        #region ==== Bookmarks
        private void addBmBtn_Click(object sender, EventArgs e)
        {
            ListViewItem bmItem = bookmarkView.Items.Add("<name>" + bookmarkView.Items.Count.ToString());
            bmItem.SubItems.Add("<description>");

            // Get copy of current Setting item, update its name and description.
            SettingItem sItem = (SettingItem)this.settings.item.Clone();
            sItem.name = bmItem.SubItems[0].Text;
            sItem.desc = bmItem.SubItems[1].Text;

            // Store setting in bookmark list.
            if (settings.bookmarks == null)
                settings.bookmarks = new List<SettingItem>();
            settings.bookmarks.Add(sItem);

            EditBookmark(bmItem);
        }

        private SettingItem FindBm(ListViewItem bmItem)
        {
            if (settings.bookmarks != null)
            {
                foreach (SettingItem sitem in this.settings.bookmarks)
                {
                    if (sitem.name == bmItem.SubItems[0].Text &&
                        sitem.desc == bmItem.SubItems[1].Text)
                    {
                        return sitem;
                    }
                }
            }

            return null;
        }

        private void EditBookmark(ListViewItem bmItem)
        {
            SettingItem sItem = FindBm(bmItem);
           
            if (sItem != null)
            {
                ListViewFieldEdit(bookmarkView, bmItem, 0);
                ListViewFieldEdit(bookmarkView, bmItem, 1);
                
                sItem.name = bmItem.SubItems[0].Text;
                sItem.desc = bmItem.SubItems[1].Text;
            }
        }

        private bool ListViewFieldEdit(ListView listView, ListViewItem item, int subCol)
        {
            Control c = listView;
            ListViewItem.ListViewSubItem subItem = item.SubItems[subCol];

            FieldBox fieldBox = new FieldBox();
            fieldBox.FieldText = subItem.Text;
            fieldBox.Location = c.PointToScreen(subItem.Bounds.Location);
            fieldBox.Show();
            fieldBox.Width = item.Bounds.Width;  // reset width
            fieldBox.Visible = false;

            if (fieldBox.ShowDialog() == DialogResult.OK)
            {
                subItem.Text = fieldBox.FieldText;
                return true;
            }
            return false;
        }

        private void deleteBmBtn_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem bmItem in bookmarkView.SelectedItems)
            {
                SettingItem sItem = FindBm(bmItem);
                if (sItem != null)
                {
                    this.settings.bookmarks.Remove(sItem);
                }
                else
                    MessageBox.Show("Failed to locate and delete bookmark information");

                bookmarkView.Items.Remove(bmItem);
            }
        }

        private void bookmarkView_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            if (e.IsSelected)
            {
                SettingItem sItem = FindBm(e.Item);
                if (sItem != null)
                {
                    settings.item = sItem;
                    LoadSettingItem();
                }
                else
                    MessageBox.Show("Failed to locate bookmark information");
            }
        }

        #endregion

        #region ==== Settings Tab
        private void axisBtn_CheckedChanged(object sender, EventArgs e)
        {
            if (noAxisBtn.Checked)
                vAxis.AxisSettings.AxisLayout = 
                    hAxis.AxisSettings.AxisLayout = GraphAxis.Layout.None;
            else if (L_axisBtn.Checked)
                vAxis.AxisSettings.AxisLayout = 
                    hAxis.AxisSettings.AxisLayout = GraphAxis.Layout.LeftBottom;
            else if (plusAxisBtn.Checked)
                vAxis.AxisSettings.AxisLayout = 
                    hAxis.AxisSettings.AxisLayout = GraphAxis.Layout.Center;
            else if (zeroAxisBtn.Checked)
                vAxis.AxisSettings.AxisLayout = 
                    hAxis.AxisSettings.AxisLayout = GraphAxis.Layout.Zero;

            DrawPoints();
        }

        private void SetAxisCheck(GraphAxis.Layout axisLayout)
        {
            doDraw = false;
            noAxisBtn.Checked = L_axisBtn.Checked = plusAxisBtn.Checked = zeroAxisBtn.Checked = false;

            switch (axisLayout)
            {
                case GraphAxis.Layout.None:
                    noAxisBtn.Checked = true;
                    break;
                case GraphAxis.Layout.LeftBottom:
                    L_axisBtn.Checked = true;
                    break;
                case GraphAxis.Layout.Center:
                    plusAxisBtn.Checked = true;
                    break;
                case GraphAxis.Layout.Zero:
                    zeroAxisBtn.Checked = true;
                    break;
            }
            doDraw = true;
        }

        private void colorList_MouseUp(object sender, MouseEventArgs e)
        {
            ListViewHitTestInfo hitInfo = colorList.HitTest(e.Location);
            if (hitInfo != null && hitInfo.Item != null)
            {
                this.colorDialog.Color = hitInfo.Item.SubItems[1].BackColor;
                if (this.colorDialog.ShowDialog() == DialogResult.OK)
                {
                    settings.item.colors[hitInfo.Item.Index] = hitInfo.Item.SubItems[1].BackColor = this.colorDialog.Color;
                }
            }
        }

        private void axisPropGrid_PropertyValueChanged(object s, PropertyValueChangedEventArgs e)
        {
            DrawPoints();
        }

        private void pointViewCk_CheckedChanged(object sender, EventArgs e)
        {
            if (pointViewCk.Checked)
                this.tabs.TabPages.Add(tabPoints);
            else
                this.tabs.TabPages.Remove(tabPoints);
        }
        #endregion

        /// <summary>
        /// Column header click fires sort.
        /// Supports two tier sorting.
        /// </summary>
        private void ColumnSort_Click(object sender, ColumnClickEventArgs e)
        {
            bool isCongtrol = System.Windows.Forms.Control.ModifierKeys == Keys.Control;

            ListView listView = sender as ListView;
            ListViewColumnSorter sorter = listView.ListViewItemSorter as ListViewColumnSorter;
            if (sorter == null)
                return;

            // Determine if clicked column is already the column that is being sorted.
            if (e.Column == sorter.SortColumn1)
            {
                // Reverse the current sort direction for this column.
                if (sorter.Order1 == SortOrder.Ascending)
                    sorter.Order1 = SortOrder.Descending;
                else
                    sorter.Order1 = SortOrder.Ascending;
            }
            else
            {
                // Set the column number that is to be sorted; default to ascending.
                sorter.SortColumn2 = sorter.SortColumn1;
                sorter.Order2 = sorter.Order1;
                sorter.SortColumn1 = e.Column;
                sorter.Order1 = SortOrder.Ascending;
            }

            // Clear old arrows and set new arrow
            foreach (ColumnHeader colHdr in listView.Columns)
                colHdr.ImageIndex = -1;

            if (sorter.SortColumn1 != sorter.SortColumn2)
                listView.Columns[sorter.SortColumn2].ImageIndex = (sorter.Order2 == SortOrder.Ascending) ? 2 : 3;

            listView.Columns[sorter.SortColumn1].ImageIndex = (sorter.Order1 == SortOrder.Ascending) ? 0 : 1;

            sorter.SortCheckedState = (isCongtrol && sorter.SortColumn1 == 0);

            sorter.SortType =
                (listView.Columns[sorter.SortColumn1].TextAlign == HorizontalAlignment.Right) ?
                ListViewColumnSorter.SortDataType.eNumeric :
                ListViewColumnSorter.SortDataType.eAlpha;

            // Perform the sort with these new sort options.
            if (listView != null)
                listView.Sort();
        }

       
    }
}
